# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Neural Network Primitives."""


from typing import Union, Dict, Callable, Tuple, Optional

import functools

from jraph import GraphsTuple
import jraph

from jax import vmap
from jax.nn import initializers
import jax.numpy as jnp

from jax_md import util
from jax_md import energy
from jax_md import partition

from ml_collections import ConfigDict

import flax.linen as nn


Array = util.Array
FeaturizerFn = Callable[[GraphsTuple, Array, Array, Optional[Array]], GraphsTuple]

f32 = jnp.float32

partial = functools.partial
normal = lambda var: initializers.variance_scaling(var, 'fan_in', 'normal')


# Nonlinearities:


UnaryFn = Callable[[Array], Array]


class BetaSwish(nn.Module):
  @nn.compact
  def __call__(self, x):
    features = x.shape[-1]
    beta = self.param('Beta', nn.initializers.ones, (features,))
    return x * nn.sigmoid(beta * x)


NONLINEARITY = {
    'none': lambda x: x,
    'relu': nn.relu,
    'swish': lambda x: BetaSwish()(x),
    'raw_swish': nn.swish,
    'tanh': nn.tanh,
    'sigmoid': nn.sigmoid,
    'silu': nn.silu,
}


def get_nonlinearity_by_name(name: str) -> UnaryFn:
  if name in NONLINEARITY:
    return NONLINEARITY[name]
  raise ValueError(f'Nonlinearity "{name}" not found.')


# Fully-Connected Networks


class MLP(nn.Module):
  features: Tuple[int, ...]
  nonlinearity: str

  use_bias: bool = True
  scalar_mlp_std: Optional[float] = None

  @nn.compact
  def __call__(self, x):
    features = self.features

    dense = partial(nn.Dense, use_bias=self.use_bias)
    phi = get_nonlinearity_by_name(self.nonlinearity)

    kernel_init = normal(self.scalar_mlp_std)

    for h in features[:-1]:
      x = phi(dense(h, kernel_init=kernel_init)(x))

    return dense(features[-1], kernel_init=normal(1.0))(x)


def mlp(hidden_features: Union[int, Tuple[int, ...]],
        nonlinearity: str,
        **kwargs) -> Callable[..., Array]:
  if isinstance(hidden_features, int):
    hidden_features = (hidden_features,)

  def mlp_fn(*args):
    fn = MLP(hidden_features, nonlinearity, **kwargs)
    return jraph.concatenated_args(fn)(*args)
  return mlp_fn


# Featurizers


def neighbor_list_featurizer(displacement_fn,
                             atom_feature_fn,
                             edge_feature_fn,
                             global_feature_fn):
  def featurize(atoms, position, neighbor, **kwargs):
    N = position.shape[0]
    graph = partition.to_jraph(neighbor, nodes=atoms)
    mask = partition.neighbor_list_mask(neighbor, True)

    Rb = position[graph.senders]
    Ra = position[graph.receivers]

    d = vmap(partial(displacement_fn, **kwargs))
    dR = d(Ra, Rb)
    dR = jnp.where(mask[:, None], dR, 1)

    return graph._replace(nodes=atom_feature_fn(graph),
                          edges=edge_feature_fn(dR))
  return featurize


# Bessel Functions


# Similar to the original Behler-Parinello features. Used by Nequip [1] and
# Schnet [2] to encode distance information.


def bessel(r_c, frequencies, r):
  rp = jnp.where(r > f32(1e-5), r, f32(1000.0))
  b = 2 / r_c * jnp.sin(frequencies * rp / r_c) / rp
  return jnp.where(r > f32(1e-5), b, 0)


class BesselEmbedding(nn.Module):
  count: int
  inner_cutoff: float
  outer_cutoff: float

  @nn.compact
  def __call__(self, rs: Array) -> Array:
    def init_fn(key, shape):
      del key
      assert len(shape) == 1
      n = shape[0]
      return jnp.arange(1, n + 1) * jnp.pi
    frequencies = self.param('frequencies', init_fn, (self.count,))
    bessel_fn = partial(bessel, self.outer_cutoff, frequencies)
    bessel_fn = vmap(energy.multiplicative_isotropic_cutoff(bessel_fn,
                                                            self.inner_cutoff,
                                                            self.outer_cutoff))
    return bessel_fn(rs)


# Scale and Shifts

# TODO(schsam): Currently, we refer to a global dataset-level scale / shift
# if no scale / shift is specified by the config. At some point, it would be
# nice to remove this.

DATASET_SHIFT_SCALE = {
    'harder_silicon': (2.2548, 0.8825)
}


def get_shift_and_scale(cfg: ConfigDict) -> Tuple[float, float]:
  if hasattr(cfg, 'scale') and hasattr(cfg, 'shift'):
    return cfg.shift, cfg.scale
  elif hasattr(cfg, 'train_dataset'):
    return DATASET_SHIFT_SCALE[cfg.train_dataset[0]]
  else:
    raise ValueError()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Dict, Union, Tuple


import functools

import e3nn_jax as e3nn

import flax.linen as nn

import jax
import jax.numpy as jnp
from jax import tree_util
from jax import vmap

from jax_md import space
from jax_md import util

import jraph

from . import util as nn_util

from .e3nn_layer import FullyConnectedTensorProductE3nn
from .e3nn_layer import Linear

import operator

from ml_collections import ConfigDict


# Types


Irreps = e3nn.Irreps
IrrepsArray = e3nn.IrrepsArray
Irrep = e3nn.Irrep
Array = util.Array
FeaturizerFn = nn_util.FeaturizerFn


get_nonlinearity_by_name = nn_util.get_nonlinearity_by_name
partial = functools.partial


# Code


def featurizer(sh_irreps: str)-> FeaturizerFn:
  """Node, edge, global features for NequIP."""
  def node_feature_fn(graph):
    """ NequIP node features are simply the input graph's nodes. """
    return graph.nodes

  def edge_feature_fn(dR):
    """ NequIP edge features.

     Builds the following:
     - the relative interatomic positions \vec{r}_{ij}
     - interatomic distances r_{ij}
     - normalized spherical harmonics Y(\hat{r}_{ij})
     """
    features = []
    edge_sh_irreps = Irreps(sh_irreps)
    edge_sh = e3nn.spherical_harmonics(edge_sh_irreps, dR, normalize=True)
    return dR, space.distance(dR), edge_sh

  def global_feature_fn(g):
    """ NequIP does not use global features. """
    return g

  return node_feature_fn, edge_feature_fn, global_feature_fn


def prod(xs):
  """From e3nn_jax/util/__init__.py."""
  return functools.reduce(operator.mul, xs, 1)


def tp_path_exists(arg_in1, arg_in2, arg_out):
  """Check if a tensor product path is viable.

  This helper function is similar to the one used in:
  https://github.com/e3nn/e3nn
  """
  arg_in1 = Irreps(arg_in1).simplify()
  arg_in2 = Irreps(arg_in2).simplify()
  arg_out = Irrep(arg_out)

  for multiplicity_1, irreps_1 in arg_in1:
    for multiplicity_2, irreps_2 in arg_in2:
      if arg_out in irreps_1 * irreps_2:
        return True
  return False


class NequIPConvolution(nn.Module):
  """NequIP Convolution.

  Implementation follows the original paper by Batzner et al.

  nature.com/articles/s41467-022-29939-5 and partially
  https://github.com/mir-group/nequip.

  Args:
        hidden_irreps: irreducible representation of hidden/latent features
        use_sc: use self-connection in network (recommended)
        nonlinearities: nonlinearities to use for even/odd irreps
        radial_net_nonlinearity: nonlinearity to use in radial MLP
        radial_net_n_hidden: number of hidden neurons in radial MLP
        radial_net_n_layers: number of hidden layers for radial MLP
        num_basis: number of Bessel basis functions to use
        n_neighbors: constant number of per-atom neighbors, used for internal
        normalization
        scalar_mlp_std: standard deviation of weight init of radial MLP

    Returns:
        Updated node features h after the convolution.
  """
  hidden_irreps: Irreps
  use_sc: bool
  nonlinearities: Union[str, Dict[str, str]]
  radial_net_nonlinearity: str = 'raw_swish'
  radial_net_n_hidden: int = 64
  radial_net_n_layers: int = 2
  num_basis: int = 8
  n_neighbors: float = 1.
  scalar_mlp_std: float = 4.0

  @nn.compact
  def __call__(
      self,
      node_features: IrrepsArray,
      node_attributes: IrrepsArray,
      edge_sh: Array,
      edge_src: Array,
      edge_dst: Array,
      edge_embedded: Array
      ) -> IrrepsArray:
    # Convolution outline in NequIP is:
    # Linear on nodes
    # TP + aggregate
    # divide by average number of neighbors
    # Concatenation
    # Linear on nodes
    # Self-connection
    # Gate
    irreps_scalars = []
    irreps_nonscalars = []
    irreps_gate_scalars = []

    # get scalar target irreps
    for multiplicity, irrep in self.hidden_irreps:
      # need the additional Irrep() here for the build, even though irrep is
      # already of type Irrep()
      if (Irrep(irrep).l == 0 and
          tp_path_exists(node_features.irreps, edge_sh.irreps, irrep)):
        irreps_scalars += [(multiplicity, irrep)]

    irreps_scalars = Irreps(irreps_scalars)

    # get non-scalar target irreps
    for multiplicity, irrep in self.hidden_irreps:
      # need the additional Irrep() here for the build, even though irrep is
      # already of type Irrep()
      if (Irrep(irrep).l > 0 and
          tp_path_exists(node_features.irreps, edge_sh.irreps, irrep)):
        irreps_nonscalars += [(multiplicity, irrep)]

    irreps_nonscalars = Irreps(irreps_nonscalars)

    # get gate scalar irreps
    if tp_path_exists(node_features.irreps, edge_sh.irreps, '0e'):
      gate_scalar_irreps_type = '0e'
    else:
      gate_scalar_irreps_type = '0o'

    for multiplicity, irreps in irreps_nonscalars:
      irreps_gate_scalars += [(multiplicity, gate_scalar_irreps_type)]

    irreps_gate_scalars = Irreps(irreps_gate_scalars)

    # final layer output irreps are all three
    # note that this order is assumed by the gate function later, i.e.
    # scalars left, then gate scalar, then non-scalars
    h_out_irreps = irreps_scalars + irreps_gate_scalars + irreps_nonscalars

    # self-connection: TP between node features and node attributes
    # this can equivalently be seen as a matrix multiplication acting on
    # the node features where the weight matrix is indexed by the node
    # attributes (typically the chemical species), i.e. a linear transform
    # that is a function of the species of the central atom
    if self.use_sc:
      self_connection = FullyConnectedTensorProductE3nn(
          h_out_irreps,
          # node_features.irreps,
          # node_attributes.irreps
          )(node_features, node_attributes)

    h = node_features

    # first linear, stays in current h-space
    h = Linear(node_features.irreps)(h)

    # map node features onto edges for tp
    edge_features = jax.tree_map(lambda x: x[edge_src], h)

    # we gather the instructions for the tp as well as the tp output irreps
    mode = 'uvu'
    trainable = 'True'
    irreps_after_tp = []
    instructions = []

    # iterate over both arguments, i.e. node irreps and edge irreps
    # if they have a valid TP path for any of the target irreps,
    # add to instructions and put in appropriate position
    # we use uvu mode (where v is a single-element sum) and weights will
    # be provide externally by the scalar MLP
    # this triple for loop is similar to the one used in e3nn and nequip
    for i, (mul_in1, irreps_in1) in enumerate(node_features.irreps):
      for j, (_, irreps_in2) in enumerate(edge_sh.irreps):
        for curr_irreps_out in irreps_in1 * irreps_in2:
          if curr_irreps_out in h_out_irreps:
            k = len(irreps_after_tp)
            irreps_after_tp += [(mul_in1, curr_irreps_out)]
            instructions += [(i, j, k, mode, trainable)]

    # we will likely have constructed irreps in a non-l-increasing order
    # so we sort them to be in a l-increasing order
    irreps_after_tp, p, _ = Irreps(irreps_after_tp).sort()

    # if we sort the target irreps, we will have to sort the instructions
    # acoordingly, using the permutation indices
    sorted_instructions = []

    for irreps_in1, irreps_in2, irreps_out, mode, trainable in instructions:
      sorted_instructions += [
          (irreps_in1, irreps_in2, p[irreps_out], mode, trainable)]

    # TP between spherical harmonics embedding of the edge vector
    # Y_ij(\hat{r}) and neighboring node h_j, weighted on a per-element basis
    # by the radial network R(r_ij)
    tp = e3nn.FunctionalTensorProduct(
        irreps_in1=edge_features.irreps,
        irreps_in2=edge_sh.irreps,
        irreps_out=irreps_after_tp,
        instructions=sorted_instructions
    )

    # scalar radial network, number of output neurons is the total number of
    # tensor product paths, nonlinearity must have f(0)=0 and MLP must not
    # have biases
    n_tp_weights = 0

    # get output dim of radial MLP / number of TP weights
    for ins in tp.instructions:
      if ins.has_weight:
        n_tp_weights += prod(ins.path_shape)

    # build radial MLP R(r) that maps from interatomic distances to TP weights
    # must not use bias to that R(0)=0
    fc = nn_util.MLP(
        (self.radial_net_n_hidden,) * self.radial_net_n_layers + (n_tp_weights,),
        self.radial_net_nonlinearity,
        use_bias=False,
        scalar_mlp_std=self.scalar_mlp_std
        )

    # the TP weights (v dimension) are given by the FC
    weight = fc(edge_embedded)

    # tp between node features that have been mapped onto edges and edge RSH
    # weighted by FC weight, we vmap over the dimension of the edges
    edge_features = jax.vmap(tp.left_right)(weight, edge_features, edge_sh)

    # aggregate edges onto nodes after tp using e3nn-jax's index_add
    h = jax.tree_map(
        lambda x: e3nn.index_add(edge_dst, x, out_dim=h.shape[0]),
        edge_features
        )

    # normalize by the average (not local) number of neighbors
    h = h / self.n_neighbors

    # second linear, now we create extra gate scalars by mapping to h-out
    h = Linear(h_out_irreps)(h)

    # self-connection, similar to a resnet-update that sums the output from
    # the TP to chemistry-weighted h
    if self.use_sc:
      h = h + self_connection

    # gate nonlinearity, applied to gate data, consisting of:
    # a) regular scalars,
    # b) gate scalars, and
    # c) non-scalars to be gated
    # in this order
    gate_fn = partial(
        e3nn.gate,
        even_act=get_nonlinearity_by_name(self.nonlinearities['e']),
        odd_act=get_nonlinearity_by_name(self.nonlinearities['o']),
        even_gate_act=get_nonlinearity_by_name(self.nonlinearities['e']),
        odd_gate_act=get_nonlinearity_by_name(self.nonlinearities['o'])
        )

    h = gate_fn(h)

    return h


class NequIPEnergyModel(nn.Module):
  """NequIP.

  Implementation follows the original paper by Batzner et al.

  nature.com/articles/s41467-022-29939-5 and partially
  https://github.com/mir-group/nequip.

    Args:
        graph_net_steps: number of NequIP convolutional layers
        use_sc: use self-connection in network (recommended)
        nonlinearities: nonlinearities to use for even/odd irreps
        n_element: number of chemical elements in input data
        hidden_irreps: irreducible representation of hidden/latent features
        sh_irreps: irreducible representations on the edges
        num_basis: number of Bessel basis functions to use
        r_max: radial cutoff used in length units
        radial_net_nonlinearity: nonlinearity to use in radial MLP
        radial_net_n_hidden: number of hidden neurons in radial MLP
        radial_net_n_layers: number of hidden layers for radial MLP
        shift: per-atom energy shift
        scale: per-atom energy scale
        n_neighbors: constant number of per-atom neighbors, used for internal
        normalization
        scalar_mlp_std: standard deviation of weight init of radial MLP

    Returns:
        Potential energy of the inputs.
  """

  graph_net_steps: int
  use_sc: bool
  nonlinearities: Union[str, Dict[str, str]]
  n_elements: int

  hidden_irreps: str
  sh_irreps: str

  num_basis: int = 8
  r_max: float = 4.

  radial_net_nonlinearity: str = 'raw_swish'
  radial_net_n_hidden: int = 64
  radial_net_n_layers: int = 2

  shift: float = 0.
  scale: float = 1.
  n_neighbors: float = 1.
  scalar_mlp_std: float = 4.0

  @nn.compact
  def __call__(self, graph):
    r_max = jnp.float32(self.r_max)
    hidden_irreps = Irreps(self.hidden_irreps)

    # get src/dst from graph
    edge_src = graph.senders
    edge_dst = graph.receivers

    # node features
    embedding_irreps = Irreps(f'{self.n_elements}x0e')
    node_attrs = IrrepsArray(embedding_irreps, graph.nodes)

    # edge embedding
    _, scalar_dr_edge, edge_sh = graph.edges

    embedded_dr_edge = nn_util.BesselEmbedding(
        count=self.num_basis,
        inner_cutoff=r_max - 0.5,
        outer_cutoff=r_max
        )(scalar_dr_edge)

    # embedding layer
    h_node = Linear(irreps_out=Irreps(hidden_irreps))(node_attrs)

    # convolutions
    for _ in range(self.graph_net_steps):
      h_node = NequIPConvolution(
          hidden_irreps=hidden_irreps,
          use_sc=self.use_sc,
          nonlinearities=self.nonlinearities,
          radial_net_nonlinearity=self.radial_net_nonlinearity,
          radial_net_n_hidden=self.radial_net_n_hidden,
          radial_net_n_layers=self.radial_net_n_layers,
          num_basis=self.num_basis,
          n_neighbors=self.n_neighbors,
          scalar_mlp_std=self.scalar_mlp_std
          )(h_node,
            node_attrs,
            edge_sh,
            edge_src,
            edge_dst,
            embedded_dr_edge
            )

    # output block, two Linears that decay dimensions from h to h//2 to 1
    for mul, ir in h_node.irreps:
      if ir == Irrep('0e'):
        mul_second_to_final = mul // 2

    second_to_final_irreps = Irreps(f'{mul_second_to_final}x0e')
    final_irreps = Irreps('1x0e')

    h_node = Linear(irreps_out=second_to_final_irreps)(h_node)
    atomic_output = Linear(irreps_out=final_irreps)(h_node).array

    # shift + scale atomic energies
    atomic_output = self.scale * atomic_output + self.shift

    # this aggregation follows jraph/_src/models.py
    n_graph = graph.n_node.shape[0]
    graph_idx = jnp.arange(n_graph)
    sum_n_node = tree_util.tree_leaves(graph.nodes)[0].shape[0]
    node_gr_idx = jnp.repeat(
        graph_idx,
        graph.n_node,
        axis=0,
        total_repeat_length=sum_n_node
        )

    global_output = tree_util.tree_map(
        lambda n: jraph.segment_sum(
            n,
            node_gr_idx,
            n_graph
            ), atomic_output)

    return global_output


def model_from_config(cfg: ConfigDict
                      ) -> Tuple[FeaturizerFn, NequIPEnergyModel]:
  """Model replication of NequIP.

  Implementation follows the original paper by Batzner et al.

  nature.com/articles/s41467-022-29939-5 and partially
  https://github.com/mir-group/nequip.
  """
  shift, scale = nn_util.get_shift_and_scale(cfg)

  model = NequIPEnergyModel(
      graph_net_steps=cfg.graph_net_steps,
      use_sc=cfg.use_sc,
      nonlinearities=cfg.nonlinearities,
      n_elements=cfg.n_elements,
      hidden_irreps=cfg.hidden_irreps,
      sh_irreps=cfg.sh_irreps,
      num_basis=cfg.num_basis,
      r_max=cfg.r_max,
      radial_net_nonlinearity=cfg.radial_net_nonlinearity,
      radial_net_n_hidden=cfg.radial_net_n_hidden,
      radial_net_n_layers=cfg.radial_net_n_layers,
      shift=shift,
      scale=scale,
      n_neighbors=cfg.n_neighbors,
      scalar_mlp_std=cfg.scalar_mlp_std,
      )

  return featurizer(cfg.sh_irreps), model
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Dict, Union, Tuple


import functools

import e3nn_jax as e3nn

import flax.linen as nn

import jax
import jax.numpy as jnp
from jax import tree_util
from jax import vmap

from jax_md import space
from jax_md import util

import jraph

from . import util as nn_util

from .e3nn_layer import FullyConnectedTensorProductE3nn
from .e3nn_layer import Linear

import operator

from ml_collections import ConfigDict


# Types


Irreps = e3nn.Irreps
IrrepsArray = e3nn.IrrepsArray
Irrep = e3nn.Irrep
Array = util.Array
FeaturizerFn = nn_util.FeaturizerFn


get_nonlinearity_by_name = nn_util.get_nonlinearity_by_name
partial = functools.partial


# Code


def featurizer(sh_irreps: str)-> FeaturizerFn:
  """Node, edge, global features for NequIP."""
  def node_feature_fn(graph):
    """ NequIP node features are simply the input graph's nodes. """
    return graph.nodes

  def edge_feature_fn(dR):
    """ NequIP edge features.

     Builds the following:
     - the relative interatomic positions \vec{r}_{ij}
     - interatomic distances r_{ij}
     - normalized spherical harmonics Y(\hat{r}_{ij})
     """
    features = []
    edge_sh_irreps = Irreps(sh_irreps)
    edge_sh = e3nn.spherical_harmonics(edge_sh_irreps, dR, normalize=True)
    return dR, space.distance(dR), edge_sh

  def global_feature_fn(g):
    """ NequIP does not use global features. """
    return g

  return node_feature_fn, edge_feature_fn, global_feature_fn


def prod(xs):
  """From e3nn_jax/util/__init__.py."""
  return functools.reduce(operator.mul, xs, 1)


def tp_path_exists(arg_in1, arg_in2, arg_out):
  """Check if a tensor product path is viable.

  This helper function is similar to the one used in:
  https://github.com/e3nn/e3nn
  """
  arg_in1 = Irreps(arg_in1).simplify()
  arg_in2 = Irreps(arg_in2).simplify()
  arg_out = Irrep(arg_out)

  for multiplicity_1, irreps_1 in arg_in1:
    for multiplicity_2, irreps_2 in arg_in2:
      if arg_out in irreps_1 * irreps_2:
        return True
  return False


class NequIPConvolution(nn.Module):
  """NequIP Convolution.

  Implementation follows the original paper by Batzner et al.

  nature.com/articles/s41467-022-29939-5 and partially
  https://github.com/mir-group/nequip.

  Args:
        hidden_irreps: irreducible representation of hidden/latent features
        use_sc: use self-connection in network (recommended)
        nonlinearities: nonlinearities to use for even/odd irreps
        radial_net_nonlinearity: nonlinearity to use in radial MLP
        radial_net_n_hidden: number of hidden neurons in radial MLP
        radial_net_n_layers: number of hidden layers for radial MLP
        num_basis: number of Bessel basis functions to use
        n_neighbors: constant number of per-atom neighbors, used for internal
        normalization
        scalar_mlp_std: standard deviation of weight init of radial MLP

    Returns:
        Updated node features h after the convolution.
  """
  hidden_irreps: Irreps
  use_sc: bool
  nonlinearities: Union[str, Dict[str, str]]
  radial_net_nonlinearity: str = 'raw_swish'
  radial_net_n_hidden: int = 64
  radial_net_n_layers: int = 2
  num_basis: int = 8
  n_neighbors: float = 1.
  scalar_mlp_std: float = 4.0

  @nn.compact
  def __call__(
      self,
      node_features: IrrepsArray,
      node_attributes: IrrepsArray,
      edge_sh: Array,
      edge_src: Array,
      edge_dst: Array,
      edge_embedded: Array
      ) -> IrrepsArray:
    # Convolution outline in NequIP is:
    # Linear on nodes
    # TP + aggregate
    # divide by average number of neighbors
    # Concatenation
    # Linear on nodes
    # Self-connection
    # Gate
    irreps_scalars = []
    irreps_nonscalars = []
    irreps_gate_scalars = []

    # get scalar target irreps
    for multiplicity, irrep in self.hidden_irreps:
      # need the additional Irrep() here for the build, even though irrep is
      # already of type Irrep()
      if (Irrep(irrep).l == 0 and
          tp_path_exists(node_features.irreps, edge_sh.irreps, irrep)):
        irreps_scalars += [(multiplicity, irrep)]

    irreps_scalars = Irreps(irreps_scalars)

    # get non-scalar target irreps
    for multiplicity, irrep in self.hidden_irreps:
      # need the additional Irrep() here for the build, even though irrep is
      # already of type Irrep()
      if (Irrep(irrep).l > 0 and
          tp_path_exists(node_features.irreps, edge_sh.irreps, irrep)):
        irreps_nonscalars += [(multiplicity, irrep)]

    irreps_nonscalars = Irreps(irreps_nonscalars)

    # get gate scalar irreps
    if tp_path_exists(node_features.irreps, edge_sh.irreps, '0e'):
      gate_scalar_irreps_type = '0e'
    else:
      gate_scalar_irreps_type = '0o'

    for multiplicity, irreps in irreps_nonscalars:
      irreps_gate_scalars += [(multiplicity, gate_scalar_irreps_type)]

    irreps_gate_scalars = Irreps(irreps_gate_scalars)

    # final layer output irreps are all three
    # note that this order is assumed by the gate function later, i.e.
    # scalars left, then gate scalar, then non-scalars
    h_out_irreps = irreps_scalars + irreps_gate_scalars + irreps_nonscalars

    # self-connection: TP between node features and node attributes
    # this can equivalently be seen as a matrix multiplication acting on
    # the node features where the weight matrix is indexed by the node
    # attributes (typically the chemical species), i.e. a linear transform
    # that is a function of the species of the central atom
    if self.use_sc:
      self_connection = FullyConnectedTensorProductE3nn(
          h_out_irreps,
          # node_features.irreps,
          # node_attributes.irreps
          )(node_features, node_attributes)

    h = node_features

    # first linear, stays in current h-space
    h = Linear(node_features.irreps)(h)

    # map node features onto edges for tp
    edge_features = jax.tree_map(lambda x: x[edge_src], h)

    # we gather the instructions for the tp as well as the tp output irreps
    mode = 'uvu'
    trainable = 'True'
    irreps_after_tp = []
    instructions = []

    # iterate over both arguments, i.e. node irreps and edge irreps
    # if they have a valid TP path for any of the target irreps,
    # add to instructions and put in appropriate position
    # we use uvu mode (where v is a single-element sum) and weights will
    # be provide externally by the scalar MLP
    # this triple for loop is similar to the one used in e3nn and nequip
    for i, (mul_in1, irreps_in1) in enumerate(node_features.irreps):
      for j, (_, irreps_in2) in enumerate(edge_sh.irreps):
        for curr_irreps_out in irreps_in1 * irreps_in2:
          if curr_irreps_out in h_out_irreps:
            k = len(irreps_after_tp)
            irreps_after_tp += [(mul_in1, curr_irreps_out)]
            instructions += [(i, j, k, mode, trainable)]

    # we will likely have constructed irreps in a non-l-increasing order
    # so we sort them to be in a l-increasing order
    irreps_after_tp, p, _ = Irreps(irreps_after_tp).sort()

    # if we sort the target irreps, we will have to sort the instructions
    # acoordingly, using the permutation indices
    sorted_instructions = []

    for irreps_in1, irreps_in2, irreps_out, mode, trainable in instructions:
      sorted_instructions += [
          (irreps_in1, irreps_in2, p[irreps_out], mode, trainable)]

    # TP between spherical harmonics embedding of the edge vector
    # Y_ij(\hat{r}) and neighboring node h_j, weighted on a per-element basis
    # by the radial network R(r_ij)
    tp = e3nn.FunctionalTensorProduct(
        irreps_in1=edge_features.irreps,
        irreps_in2=edge_sh.irreps,
        irreps_out=irreps_after_tp,
        instructions=sorted_instructions
    )

    # scalar radial network, number of output neurons is the total number of
    # tensor product paths, nonlinearity must have f(0)=0 and MLP must not
    # have biases
    n_tp_weights = 0

    # get output dim of radial MLP / number of TP weights
    for ins in tp.instructions:
      if ins.has_weight:
        n_tp_weights += prod(ins.path_shape)

    # build radial MLP R(r) that maps from interatomic distances to TP weights
    # must not use bias to that R(0)=0
    fc = nn_util.MLP(
        (self.radial_net_n_hidden,) * self.radial_net_n_layers + (n_tp_weights,),
        self.radial_net_nonlinearity,
        use_bias=False,
        scalar_mlp_std=self.scalar_mlp_std
        )

    # the TP weights (v dimension) are given by the FC
    weight = fc(edge_embedded)

    # tp between node features that have been mapped onto edges and edge RSH
    # weighted by FC weight, we vmap over the dimension of the edges
    edge_features = jax.vmap(tp.left_right)(weight, edge_features, edge_sh)
    # TODO: It's not great that e3nn_jax automatically upcasts internally,
    # but this would need to be fixed at the e3nn level.
    edge_features = jax.tree_map(lambda x: x.astype(h.dtype), edge_features)

    # aggregate edges onto nodes after tp using e3nn-jax's index_add
    h_type = h.dtype
    h = jax.tree_map(
        lambda x: e3nn.index_add(edge_dst, x, out_dim=h.shape[0]),
        edge_features
        )
    # TODO: Remove this once e3nn_jax doesn't upcast inputs.
    h = jax.tree_map(lambda x: x.astype(h_type), h)

    # normalize by the average (not local) number of neighbors
    h = h / self.n_neighbors

    # second linear, now we create extra gate scalars by mapping to h-out
    h = Linear(h_out_irreps)(h)

    # self-connection, similar to a resnet-update that sums the output from
    # the TP to chemistry-weighted h
    if self.use_sc:
      h = h + self_connection

    # gate nonlinearity, applied to gate data, consisting of:
    # a) regular scalars,
    # b) gate scalars, and
    # c) non-scalars to be gated
    # in this order
    gate_fn = partial(
        e3nn.gate,
        even_act=get_nonlinearity_by_name(self.nonlinearities['e']),
        odd_act=get_nonlinearity_by_name(self.nonlinearities['o']),
        even_gate_act=get_nonlinearity_by_name(self.nonlinearities['e']),
        odd_gate_act=get_nonlinearity_by_name(self.nonlinearities['o'])
        )

    h = gate_fn(h)
    # TODO: Remove this once e3nn_jax doesn't upcast inputs.
    h = jax.tree_map(lambda x: x.astype(h_type), h)

    return h


class NequIPEnergyModel(nn.Module):
  """NequIP.

  Implementation follows the original paper by Batzner et al.

  nature.com/articles/s41467-022-29939-5 and partially
  https://github.com/mir-group/nequip.

    Args:
        graph_net_steps: number of NequIP convolutional layers
        use_sc: use self-connection in network (recommended)
        nonlinearities: nonlinearities to use for even/odd irreps
        n_element: number of chemical elements in input data
        hidden_irreps: irreducible representation of hidden/latent features
        sh_irreps: irreducible representations on the edges
        num_basis: number of Bessel basis functions to use
        r_max: radial cutoff used in length units
        radial_net_nonlinearity: nonlinearity to use in radial MLP
        radial_net_n_hidden: number of hidden neurons in radial MLP
        radial_net_n_layers: number of hidden layers for radial MLP
        shift: per-atom energy shift
        scale: per-atom energy scale
        n_neighbors: constant number of per-atom neighbors, used for internal
        normalization
        scalar_mlp_std: standard deviation of weight init of radial MLP

    Returns:
        Potential energy of the inputs.
  """

  graph_net_steps: int
  use_sc: bool
  nonlinearities: Union[str, Dict[str, str]]
  n_elements: int

  hidden_irreps: str
  sh_irreps: str

  num_basis: int = 8
  r_max: float = 4.

  radial_net_nonlinearity: str = 'raw_swish'
  radial_net_n_hidden: int = 64
  radial_net_n_layers: int = 2

  shift: float = 0.
  scale: float = 1.
  n_neighbors: float = 1.
  scalar_mlp_std: float = 4.0

  @nn.compact
  def __call__(self, graph):
    r_max = jnp.float32(self.r_max)
    hidden_irreps = Irreps(self.hidden_irreps)

    # get src/dst from graph
    edge_src = graph.senders
    edge_dst = graph.receivers

    # node features
    embedding_irreps = Irreps(f'{self.n_elements}x0e')
    node_attrs = IrrepsArray(embedding_irreps, graph.nodes)

    # edge embedding
    _, scalar_dr_edge, edge_sh = graph.edges

    embedded_dr_edge = nn_util.BesselEmbedding(
        count=self.num_basis,
        inner_cutoff=r_max - 0.5,
        outer_cutoff=r_max
        )(scalar_dr_edge)

    # embedding layer
    h_node = Linear(irreps_out=Irreps(hidden_irreps))(node_attrs)

    # convolutions
    for _ in range(self.graph_net_steps):
      h_node = NequIPConvolution(
          hidden_irreps=hidden_irreps,
          use_sc=self.use_sc,
          nonlinearities=self.nonlinearities,
          radial_net_nonlinearity=self.radial_net_nonlinearity,
          radial_net_n_hidden=self.radial_net_n_hidden,
          radial_net_n_layers=self.radial_net_n_layers,
          num_basis=self.num_basis,
          n_neighbors=self.n_neighbors,
          scalar_mlp_std=self.scalar_mlp_std
          )(h_node,
            node_attrs,
            edge_sh,
            edge_src,
            edge_dst,
            embedded_dr_edge
            )

    # output block, two Linears that decay dimensions from h to h//2 to 1
    for mul, ir in h_node.irreps:
      if ir == Irrep('0e'):
        mul_second_to_final = mul // 2

    second_to_final_irreps = Irreps(f'{mul_second_to_final}x0e')
    final_irreps = Irreps('1x0e')

    h_node = Linear(irreps_out=second_to_final_irreps)(h_node)
    atomic_output = Linear(irreps_out=final_irreps)(h_node).array

    # shift + scale atomic energies
    atomic_output = self.scale * atomic_output + self.shift

    # this aggregation follows jraph/_src/models.py
    n_graph = graph.n_node.shape[0]
    graph_idx = jnp.arange(n_graph)
    sum_n_node = tree_util.tree_leaves(graph.nodes)[0].shape[0]
    node_gr_idx = jnp.repeat(
        graph_idx,
        graph.n_node,
        axis=0,
        total_repeat_length=sum_n_node
        )

    global_output = tree_util.tree_map(
        lambda n: jraph.segment_sum(
            n,
            node_gr_idx,
            n_graph
            ), atomic_output)

    return global_output


def model_from_config(cfg: ConfigDict
                      ) -> Tuple[FeaturizerFn, NequIPEnergyModel]:
  """Model replication of NequIP.

  Implementation follows the original paper by Batzner et al.

  nature.com/articles/s41467-022-29939-5 and partially
  https://github.com/mir-group/nequip.
  """
  shift, scale = nn_util.get_shift_and_scale(cfg)

  model = NequIPEnergyModel(
      graph_net_steps=cfg.graph_net_steps,
      use_sc=cfg.use_sc,
      nonlinearities=cfg.nonlinearities,
      n_elements=cfg.n_elements,
      hidden_irreps=cfg.hidden_irreps,
      sh_irreps=cfg.sh_irreps,
      num_basis=cfg.num_basis,
      r_max=cfg.r_max,
      radial_net_nonlinearity=cfg.radial_net_nonlinearity,
      radial_net_n_hidden=cfg.radial_net_n_hidden,
      radial_net_n_layers=cfg.radial_net_n_layers,
      shift=shift,
      scale=scale,
      n_neighbors=cfg.n_neighbors,
      scalar_mlp_std=cfg.scalar_mlp_std,
      )

  return featurizer(cfg.sh_irreps), model


def default_config() -> ConfigDict:
  config = ConfigDict()

  config.graph_net_steps = 5
  config.nonlinearities = {'e': 'raw_swish', 'o': 'tanh'}
  config.use_sc = True
  config.n_elements = 94
  config.hidden_irreps = '128x0e + 64x1e + 4x2e'
  config.sh_irreps = '1x0e + 1x1e + 1x2e'
  config.num_basis = 8
  config.r_max = 5.
  config.radial_net_nonlinearity = 'raw_swish'
  config.radial_net_n_hidden = 64
  config.radial_net_n_layers = 2

  # average number of neighbors per atom, used to divide activations are sum
  # in the nequip convolution, helpful for internal normalization.
  config.n_neighbors = 10.

  # Standard deviation used for the initializer of the weight matrix in the
  # radial scalar MLP
  config.scalar_mlp_std = 4.

  return config



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jax_md import space
from jax_md import energy
from jax_md import minimize
from jax_md import simulate
from jax_md import smap
from jax_md import partition
from jax_md import elasticity
from jax_md import dataclasses
from jax_md import nn
from jax_md import interpolate
from jax_md import util
from jax_md import io
from jax_md import rigid_body

try:
  # Attempt to load colab_tools if IPython is installed.
  from jax_md import colab_tools
except:
  pass



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import NamedTuple, Tuple

import os

from jax import ShapedArray
from jax import eval_shape
from jax import random
from jax import tree_map

import jax.numpy as jnp

from jax_md import util

import json
import jraph

import e3nn_jax as e3nn

from . import nequip

import optax

from ml_collections import ConfigDict

from flax import serialization
import flax.linen as nn


f32 = jnp.float32
i32 = jnp.int32

GraphsTuple = jraph.GraphsTuple

IrrepsArray = e3nn.IrrepsArray

NUM_ELEMENTS = 94

PyTree = util.PyTree


def model_from_config(cfg: ConfigDict) -> nn.Module:
  model_family = cfg.get('model_family', 'nequip')
  if model_family == 'nequip':
    return nequip.model_from_config(cfg)
  else:
    raise ValueError(f'Unrecognized model family: {model_family}')


def minimum_batch_size(cfg: ConfigDict) -> int:
  if not hasattr(cfg, 'train_batch_size'):
    return 1
  if isinstance(cfg.train_batch_size, int):
    return cfg.train_batch_size
  return min(cfg.train_batch_size)


class ScaleLROnPlateau(NamedTuple):
  step_size: jnp.ndarray
  minimum_loss: jnp.ndarray
  steps_without_reduction: jnp.ndarray
  max_steps_without_reduction: jnp.ndarray
  reduction_factor: jnp.ndarray


def scale_lr_on_plateau(initial_step_size: float,
                        max_steps_without_reduction: int,
                        reduction_factor: float
                        ) -> optax.GradientTransformation:
  def init_fn(params):
    del params
    return ScaleLROnPlateau(initial_step_size,
                            jnp.inf,
                            0,
                            max_steps_without_reduction,
                            reduction_factor)

  def update_fn(updates, state, params=None):
    del params
    updates = jax.tree_map(lambda g: g * state.step_size, updates)
    return updates, state

  return optax.GradientTransformation(init_fn, update_fn)


def optimizer(cfg: ConfigDict) -> optax.OptState:
  epoch_size = cfg.epoch_size if hasattr(cfg, 'epoch_size') else -1
  # TODO:
  # if epoch_size < 0:
  #   epoch_size = aggregate_dataset_size(cfg.train_dataset)
  # Maybe replace stubbed in value.

  batch_size = minimum_batch_size(cfg)
  total_steps = cfg.epochs * (epoch_size // batch_size)
  warmup_steps = cfg.get('warmup_steps', 0)

  if cfg.schedule == 'constant':
    schedule = cfg.learning_rate
  elif cfg.schedule == 'linear_decay':
    schedule = optax.polynomial_schedule(cfg.learning_rate, 0.0, 1, total_steps)
  elif cfg.schedule == 'cosine_decay':
    schedule = optax.cosine_decay_schedule(cfg.learning_rate, total_steps)
  elif cfg.schedule == 'warmup_cosine_decay':
    schedule = optax.warmup_cosine_decay_schedule(
        1e-7, cfg.learning_rate, warmup_steps, total_steps)
  elif cfg.schedule == 'scale_on_plateau':
    max_plateau_steps = cfg.max_lr_plateau_epochs // cfg.epochs_per_eval
    return optax.chain(
        optax.scale_by_adam(),
        scale_lr_on_plateau(-cfg.learning_rate, max_plateau_steps, 0.8)
        )
  else:
    raise ValueError(f'Unknown learning rate schedule, "{cfg.schedule}".')

  if not hasattr(cfg, 'l2_regularization') or cfg.l2_regularization == 0.0:
    return optax.adam(schedule)

  return optax.adamw(schedule, weight_decay=cfg.l2_regularization)


def load_model(directory: str) -> Tuple[ConfigDict, nn.Module, PyTree]:
  with open(os.path.join(directory, 'config.json'), 'r') as f:
    c = json.loads(json.loads(f.read()))
    c = ConfigDict(c)

  # Now initialize the model and the optimizer functions.
  featurizer, model = model_from_config(c)
  opt_init, _ = optimizer(c)

  # Create a dummy example to use for abstract initialization since we only need
  # the PyTree structure to deserialize.
  def make_irreps(x):
    return IrrepsArray('0e + 1e', x)

  graph = GraphsTuple(
    ShapedArray((1, NUM_ELEMENTS), f32),    # Nodes     (nodes, features)
    (ShapedArray((1, 3), f32),   # dR        (edges, spatial)
     ShapedArray((1,), f32),      # dr
     eval_shape(make_irreps, ShapedArray((1, 4), f32))),  # Spherical Harmonics
    ShapedArray((1,), i32),      # senders   (edges,)
    ShapedArray((1,), i32),      # receivers (edges,)
    ShapedArray((1, 1), f32),      # globals   (graphs,)
    ShapedArray((1,), i32),      # n_node    (graphs,)
    ShapedArray((1,), i32))     # n_edge    (graphs,)

  def init_opt_and_model(graph):
    key = random.PRNGKey(0)
    params = model.init(key, graph)
    state = opt_init(params)
    return params, state

  abstract_params, abstract_state = eval_shape(init_opt_and_model, graph)

  # Now that we have the structure, load the data using FLAX checkpointing.
  ckpt_data = (0, abstract_params, abstract_state)

  checkpoints = [c for c in os.listdir(directory) if 'checkpoint' in c]
  assert len(checkpoints) == 1

  checkpoint = os.path.join(directory, checkpoints[0])

  with open(checkpoint, 'rb') as f:
    ckpt = serialization.from_bytes(ckpt_data, f.read())

  params = tree_map(lambda x: x.astype(f32), ckpt[1])
  return c, featurizer, model, params


# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Neural Network Primitives."""

from typing import Callable, Tuple, Dict, Any, Optional

import numpy as onp

import jax
from jax import vmap, jit
import jax.numpy as jnp

from jax_md import space, dataclasses, quantity, partition, smap
from jax_md import util as jmd_util
import haiku as hk

from collections import namedtuple
from functools import partial, reduce
from jax.tree_util import tree_map
from jax import ops

import jraph

from ._nn import behler_parrinello
from ._nn import nequip
from ._nn import gnome
from ._nn import util

# Typing


Array = jmd_util.Array
f32 = jmd_util.f32
f64 = jmd_util.f64

InitFn = Callable[..., Array]
CallFn = Callable[..., Array]

DisplacementOrMetricFn = space.DisplacementOrMetricFn
DisplacementFn = space.DisplacementFn
NeighborList = partition.NeighborList

# TO BE DELETED BELOW:
# Graph neural network primitives

"""
  Our implementation here is based off the outstanding GraphNets library by
  DeepMind at, www.github.com/deepmind/graph_nets. This implementation was also
  heavily influenced by work done by Thomas Keck. We implement a subset of the
  functionality from the graph nets library to be compatible with jax-md
  states and neighbor lists, end-to-end jit compilation, and easy batching.

  Graphs are described by node states, edge states, a global state, and
  outgoing / incoming edges.

  We provide two components:

    1) A GraphIndependent layer that applies a neural network separately to the
       node states, the edge states, and the globals. This is often used as an
       encoding or decoding step.
    2) A GraphNetwork layer that transforms the nodes, edges, and globals using
       neural networks following Battaglia et al. (). Here, we use
       sum-message-aggregation.

  The graphs network components implemented here implement identical functions
  to the DeepMind library. However, to be compatible with jax-md, there are
  significant differences in the graph layout used here to the reference
  implementation. See `GraphsTuple` for details.
"""

@dataclasses.dataclass
class GraphsTuple(object):
    """A struct containing graph data.

    Attributes:
      nodes: For a graph with `N_nodes`, this is an `[N_nodes, node_dimension]`
        array containing the state of each node in the graph.
      edges: For a graph whose degree is bounded by max_degree, this is an
        `[N_nodes, max_degree, edge_dimension]`. Here `edges[i, j]` is the
        state of the outgoing edge from node `i` to node `edge_idx[i, j]`.
      globals: An array of shape `[global_dimension]`.
      edge_idx: An integer array of shape `[N_nodes, max_degree]` where
        `edge_idx[i, j]` is the id of the j-th outgoing edge from node `i`.
        Empty entries (that don't contain an edge) are denoted by
        `edge_idx[i, j] == N_nodes`.
    """
    nodes: jnp.ndarray
    edges: jnp.ndarray
    globals: jnp.ndarray
    edge_idx: jnp.ndarray

    _replace = dataclasses.replace


def concatenate_graph_features(graphs: Tuple[GraphsTuple, ...]) -> GraphsTuple:
  """Given a list of GraphsTuple returns a new concatenated GraphsTuple.

  Note that currently we do not check that the graphs have consistent edge
  connectivity.
  """
  graph = graphs[0]
  return graph._replace(
    nodes=jnp.concatenate([g.nodes for g in graphs], axis=-1),
    edges=jnp.concatenate([g.edges for g in graphs], axis=-1),
    globals=jnp.concatenate([g.globals for g in graphs], axis=-1),  # pytype: disable=missing-parameter
  )


def GraphMapFeatures(edge_fn: Callable[[Array], Array],
                     node_fn: Callable[[Array], Array],
                     global_fn: Callable[[Array], Array]
                     ) -> Callable[[GraphsTuple], GraphsTuple]:
  """Applies functions independently to the nodes, edges, and global states.
  """
  identity = lambda x: x
  _node_fn = vmap(node_fn) if node_fn is not None else identity
  _edge_fn = vmap(vmap(edge_fn)) if edge_fn is not None else identity
  _global_fn = global_fn if global_fn is not None else identity

  def embed_fn(graph):
    return dataclasses.replace(
        graph,
        nodes=_node_fn(graph.nodes),
        edges=_edge_fn(graph.edges),
        globals=_global_fn(graph.globals)
    )
  return embed_fn


def _apply_node_fn(graph: GraphsTuple,
                   node_fn: Callable[[Array,Array, Array, Array], Array]
                   ) -> Array:
  mask = graph.edge_idx < graph.nodes.shape[0]
  mask = mask[:, :, jnp.newaxis]

  if graph.edges is not None:
    # TODO: Should we also have outgoing edges?
    flat_edges = jnp.reshape(graph.edges, (-1, graph.edges.shape[-1]))
    edge_idx = jnp.reshape(graph.edge_idx, (-1,))
    incoming_edges = jax.ops.segment_sum(
        flat_edges, edge_idx, graph.nodes.shape[0] + 1)[:-1]
    outgoing_edges = jnp.sum(graph.edges * mask, axis=1)
  else:
    incoming_edges = None
    outgoing_edges = None

  if graph.globals is not None:
    _globals = jnp.broadcast_to(graph.globals[jnp.newaxis, :],
                               graph.nodes.shape[:1] + graph.globals.shape)
  else:
    _globals = None

  return node_fn(graph.nodes, incoming_edges, outgoing_edges, _globals)


def _apply_edge_fn(graph: GraphsTuple,
                   edge_fn: Callable[[Array, Array, Array, Array], Array]
                   ) -> Array:
  if graph.nodes is not None:
    incoming_nodes = graph.nodes[graph.edge_idx]
    outgoing_nodes = jnp.broadcast_to(
        graph.nodes[:, jnp.newaxis, :],
        graph.edge_idx.shape + graph.nodes.shape[-1:])
  else:
    incoming_nodes = None
    outgoing_nodes = None

  if graph.globals is not None:
    _globals = jnp.broadcast_to(graph.globals[jnp.newaxis, jnp.newaxis, :],
                               graph.edge_idx.shape + graph.globals.shape)
  else:
    _globals = None

  mask = graph.edge_idx < graph.nodes.shape[0]
  mask = mask[:, :, jnp.newaxis]
  return edge_fn(graph.edges, incoming_nodes, outgoing_nodes, _globals) * mask


def _apply_global_fn(graph: GraphsTuple,
                     global_fn: Callable[[Array, Array, Array], Array]
                     ) -> Array:
  nodes = None if graph.nodes is None else jnp.sum(graph.nodes, axis=0)

  if graph.edges is not None:
    mask = graph.edge_idx < graph.nodes.shape[0]
    mask = mask[:, :, jnp.newaxis]
    edges = jnp.sum(graph.edges * mask, axis=(0, 1))
  else:
    edges = None

  return global_fn(nodes, edges, graph.globals)


class GraphNetwork:
  """Implementation of a Graph Network.

  See https://arxiv.org/abs/1806.01261 for more details.
  """
  def __init__(self,
               edge_fn: Callable[[Array], Array],
               node_fn: Callable[[Array], Array],
               global_fn: Callable[[Array], Array]):
    self._node_fn = (None if node_fn is None else
                     partial(_apply_node_fn, node_fn=vmap(node_fn)))

    self._edge_fn = (None if edge_fn is None else
                     partial(_apply_edge_fn, edge_fn=vmap(vmap(edge_fn))))

    self._global_fn = (None if global_fn is None else
                       partial(_apply_global_fn, global_fn=global_fn))

  def __call__(self, graph: GraphsTuple) -> GraphsTuple:
    if self._edge_fn is not None:
      graph = dataclasses.replace(graph, edges=self._edge_fn(graph))

    if self._node_fn is not None:
      graph = dataclasses.replace(graph, nodes=self._node_fn(graph))

    if self._global_fn is not None:
      graph = dataclasses.replace(graph, globals=self._global_fn(graph))

    return graph


# Prefab Networks


class GraphNetEncoder(hk.Module):
  """Implements a Graph Neural Network for energy fitting.

  Based on the network used in "Unveiling the predictive power of static
  structure in glassy systems"; Bapst et al.
  (https://www.nature.com/articles/s41567-020-0842-8). This network first
  embeds edges, nodes, and global state. Then `n_recurrences` of GraphNetwork
  layers are applied. Unlike in Bapst et al. this network does not include a
  readout, which should be added separately depending on the application.

  For example, when predicting particle mobilities, one would use a decoder
  only on the node states while a model of energies would decode only the node
  states.
  """
  def __init__(self,
               n_recurrences: int,
               mlp_sizes: Tuple[int, ...],
               mlp_kwargs: Optional[Dict[str, Any]]=None,
               format: partition.NeighborListFormat=partition.Dense,
               name: str='GraphNetEncoder'):
    super(GraphNetEncoder, self).__init__(name=name)

    if mlp_kwargs is None:
      mlp_kwargs = {}

    self._n_recurrences = n_recurrences

    embedding_fn = lambda name: hk.nets.MLP(
        output_sizes=mlp_sizes,
        activate_final=True,
        name=name,
        **mlp_kwargs)

    model_fn = lambda name: lambda *args: hk.nets.MLP(
        output_sizes=mlp_sizes,
        activate_final=True,
        name=name,
        **mlp_kwargs)(jnp.concatenate(args, axis=-1))

    if format is partition.Dense:
      self._encoder = GraphMapFeatures(
        embedding_fn('EdgeEncoder'),
        embedding_fn('NodeEncoder'),
        embedding_fn('GlobalEncoder'))
      self._propagation_network = lambda: GraphNetwork(
        model_fn('EdgeFunction'),
        model_fn('NodeFunction'),
        model_fn('GlobalFunction'))
    elif format is partition.Sparse:
      self._encoder = jraph.GraphMapFeatures(
        embedding_fn('EdgeEncoder'),
        embedding_fn('NodeEncoder'),
        embedding_fn('GlobalEncoder')
      )
      self._propagation_network = lambda: jraph.GraphNetwork(
        model_fn('EdgeFunction'),
        model_fn('NodeFunction'),
        model_fn('GlobalFunction')
      )
    else:
      raise ValueError()

  def __call__(self, graph: GraphsTuple) -> GraphsTuple:
    encoded = self._encoder(graph)
    outputs = encoded

    for _ in range(self._n_recurrences):
      inputs = concatenate_graph_features((outputs, encoded))
      outputs = self._propagation_network()(inputs)

    return outputs


# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Describes different physical quantities."""


from typing import TypeVar, Callable, Union, Tuple, Optional, Any

from absl import logging

from jax import grad, vmap, eval_shape
from jax.tree_util import tree_map, tree_reduce
import jax.numpy as jnp
from jax import ops
from jax import ShapeDtypeStruct
from jax.tree_util import tree_map, tree_reduce

from jax_md import space, dataclasses, partition, util

import functools
import operator

partial = functools.partial

# Types


Array = util.Array
f32 = util.f32
f64 = util.f64

DisplacementFn = space.DisplacementFn
MetricFn = space.MetricFn
Box = space.Box

EnergyFn = Callable[..., Array]
ForceFn = Callable[..., Array]

T = TypeVar('T')
InitFn = Callable[..., T]
ApplyFn = Callable[[T], T]
Simulator = Tuple[InitFn, ApplyFn]


# Functions


def force(energy_fn: EnergyFn) -> ForceFn:
  """Computes the force as the negative gradient of an energy."""
  return grad(lambda R, *args, **kwargs: -energy_fn(R, *args, **kwargs))


def clipped_force(energy_fn: EnergyFn, max_force: float) -> ForceFn:
  force_fn = force(energy_fn)
  def wrapped_force_fn(R, *args, **kwargs):
    force = force_fn(R, *args, **kwargs)
    force_norm = jnp.linalg.norm(force, axis=-1, keepdims=True)
    return jnp.where(force_norm > max_force,
                     force / force_norm * max_force,
                     force)

  return wrapped_force_fn


def canonicalize_force(energy_or_force_fn: Union[EnergyFn, ForceFn]) -> ForceFn:
  _force_fn = None
  def force_fn(R, **kwargs):
    nonlocal _force_fn
    if _force_fn is None:
      out_shaped = eval_shape(energy_or_force_fn, R, **kwargs)
      if isinstance(out_shaped, ShapeDtypeStruct) and out_shaped.shape == ():
        _force_fn = force(energy_or_force_fn)
      else:
        # Check that the output has the right shape to be a force.
        is_valid_force = tree_reduce(
          lambda x, y: x and y,
          tree_map(lambda x, y: x.shape == y.shape, out_shaped, R),
          True
        )
        if not is_valid_force:
          raise ValueError('Provided function should be compatible with '
                           'either an energy or a force. Found a function '
                           f'whose output has shape {out_shaped}.')

        _force_fn = energy_or_force_fn
    return _force_fn(R, **kwargs)

  return force_fn


@functools.singledispatch
def count_dof(position: Array) -> int:
  util.check_custom_simulation_type(position)
  return tree_reduce(lambda accum, x: accum + x.size, position, 0)


def volume(dimension: int, box: Box) -> float:
  if jnp.isscalar(box) or not box.ndim:
    return box ** dimension
  elif box.ndim == 1:
    return jnp.prod(box)
  elif box.ndim == 2:
    return jnp.linalg.det(box)
  raise ValueError(('Box must be either: a scalar, a vector, or a matrix. '
                    f'Found {box}.'))


def kinetic_energy(*unused_args,
                   momentum: Array=None,
                   velocity: Array=None,
                   mass: Array=f32(1.0)
                   ) -> float:
  """Computes the kinetic energy of a system.

  To avoid ambiguity, either momentum or velocity must be passed explicitly
  as a keyword argument.

  Args:
    momentum: Array specifying the momentum of the system.
    velocity: Array specifying the velocity of the system.
    mass: Array specifying the mass of the constituents.

  Returns:
    The kinetic energy of the system.
  """
  if unused_args:
    raise ValueError('To use the kinetic energy function, you must explicitly '
                     'pass either momentum or velocity as a keyword argument.')
  if momentum is not None and velocity is not None:
    raise ValueError('To use the kinetic energy function, you must pass either'
                     ' a momentum or a velocity.')

  k = (lambda v, m: v**2 * m) if momentum is None else (lambda p, m: p**2 / m)
  q = velocity if momentum is None else momentum
  util.check_custom_simulation_type(q)

  ke = tree_map(lambda m, q: 0.5 * util.high_precision_sum(k(q, m)), mass, q)
  return tree_reduce(operator.add, ke, 0.0)


def temperature(*unused_args,
                momentum: Array=None,
                velocity: Array=None,
                mass: Array=f32(1.0)
                ) -> float:
  """Computes the temperature of a system.

  To avoid ambiguity, either momentum or velocity must be passed explicitly
  as a keyword argument.

  Args:
    momentum: Array specifying the momentum of the system.
    velocity: Array specifying the velocity of the system.
    mass: Array specifying the mass of the constituents.

  Returns:
    The temperature of the system in units of the Boltzmann constant.
  """
  if unused_args:
    raise ValueError('To use the kinetic energy function, you must explicitly '
                     'pass either momentum or velocity as a keyword argument.')
  if momentum is not None and velocity is not None:
    raise ValueError('To use the kinetic energy function, you must pass either'
                     ' a momentum or a velocity.')

  t = (lambda v, m: v**2 * m) if momentum is None else (lambda p, m: p**2 / m)
  q = velocity if momentum is None else momentum
  util.check_custom_simulation_type(q)

  dof = count_dof(q)

  kT = tree_map(lambda m, q: util.high_precision_sum(t(q, m)) / dof, mass, q)
  return tree_reduce(operator.add, kT, 0.0)


def pressure(energy_fn: EnergyFn, position: Array, box: Box,
             kinetic_energy: float=0.0, **kwargs) -> float:
  """Computes the internal pressure of a system.

  Args:
    energy_fn: A function that computes the energy of the system. This
      function must take as an argument `perturbation` which perturbs the
      box shape. Any energy function constructed using `smap` or in `energy.py`
      with a standard space will satisfy this property.
    position: An array of particle positions.
    box: A box specifying the shape of the simulation volume. Used to infer the
      volume of the unit cell.
    kinetic_energy: A float specifying the kinetic energy of the system.

  Returns:
    A float specifying the pressure of the system.
  """
  dim = position.shape[1]

  def U(eps):
    try:
      return energy_fn(position, box=box, perturbation=(1 + eps), **kwargs)
    except space.UnexpectedBoxException:
      return energy_fn(position, perturbation=(1 + eps), **kwargs)

  dUdV = grad(U)
  vol_0 = volume(dim, box)

  return 1 / (dim * vol_0) * (2 * kinetic_energy - dUdV(0.0))


def stress(energy_fn: EnergyFn, position: Array, box: Box,
           mass: Array=1.0, velocity: Optional[Array]=None, **kwargs
           ) -> Array:
  """Computes the internal stress of a system.

  Args:
    energy_fn: A function that computes the energy of the system. This
      function must take as an argument `perturbation` which perturbs the
      box shape. Any energy function constructed using `smap` or in `energy.py`
      with a standard space will satisfy this property.
    position: An array of particle positions.
    box: A box specifying the shape of the simulation volume. Used to infer the
      volume of the unit cell.
    mass: The mass of the particles; only used to compute the kinetic
      contribution if `velocity` is not `None`.
    velocity: An array of atomic velocities.

  Returns:
    A float specifying the pressure of the system.
  """
  dim = position.shape[1]

  zero = jnp.zeros((dim, dim), position.dtype)
  I = jnp.eye(dim, dtype=position.dtype)

  def U(eps):
    try:
      return energy_fn(position, box=box, perturbation=(I + eps), **kwargs)
    except space.UnexpectedBoxException:
      return energy_fn(position, perturbation=(I + eps), **kwargs)

  dUdV = grad(U)
  vol_0 = volume(dim, box)

  VxV = 0.0
  if velocity is not None:
    V = velocity
    VxV = util.high_precision_sum(mass * V[:, None, :] * V[:, :, None], axis=0)

  return 1 / vol_0 * (VxV - dUdV(zero))


def cosine_angle_between_two_vectors(dR_12: Array, dR_13: Array) -> Array:
  dr_12 = space.distance(dR_12) + 1e-7
  dr_13 = space.distance(dR_13) + 1e-7
  cos_angle = jnp.dot(dR_12, dR_13) / dr_12 / dr_13
  return jnp.clip(cos_angle, -1.0, 1.0)


def cosine_angles(dR: Array) -> Array:
  """Returns cosine of angles for all atom triplets.

  Args:
    dR: Matrix of displacements; `ndarray(shape=[num_atoms, num_neighbors,
      spatial_dim])`.

  Returns:
    Tensor of cosine of angles;
    `ndarray(shape=[num_atoms, num_neighbors, num_neighbors])`.
  """

  angles_between_all_triplets = vmap(
      vmap(vmap(cosine_angle_between_two_vectors, (0, None)), (None, 0)), 0)
  return angles_between_all_triplets(dR, dR)


def is_integer(x: Array) -> bool:
  return x.dtype == jnp.int32 or x.dtype == jnp.int64


def average_pair_correlation_results(gofr, species=None):
  """ Calculate species-based averages of pair correlations.

  Average the results of pair_correlation or pair_correlation_neighbor_list,
  appropriately taking species information into account.

  When species=None, gofr is expected to be an array of shape (N,nr), where N is
  the number of species and nr is the number of radii to be considered. The
  average is calculated over all particles, so an array of shape (nr,) is
  returned.

  When species is specified, gofr is expected to be a list of nspecies arrays,
  each of shape (N,nr), where nspecies is the number of unique species types. 
  Here, the average is carried out separately for every pair of species, so the
  returned array has shape (nspecies, nspecies, nr). 

  Args:
    gofr: array of shape (N,nr) or a list of arrays of shape (N,nr), where nr is
      the number of radii for which :math:`g(r)` is calculated.
    species: Optional. Array of shape (N,) specifying the species of each 
      particle.

  Returns:
    An array of shape (nr,) for species=None, otherwise an array of shape 
      (nspecies, nspecies, nr), where nspecies is the number of unique species.
  """
  if species is None:
    return jnp.mean(gofr, axis=0)
  species_types = jnp.unique(species) #note: this returns in sorted order
  return jnp.array([ [ jnp.mean(gofr[si][species==s], axis=0) \
      for s in species_types] for si in range(species_types.size)])


def pair_correlation(displacement_or_metric: Union[DisplacementFn, MetricFn],
                     radii: Array,
                     sigma: float,
                     species: Array = None,
                     eps: float = 1e-7,
                     compute_average: bool = False):
  """Computes the pair correlation function at a mesh of distances.

  The pair correlation function measures the number of particles at a given
  distance from a central particle. The pair correlation function is defined by

  .. math::
    g(r) = <\sum_{i \\neq j}\delta(r - |r_i - r_j|)>

  We make the approximation,

  .. math::
    \delta(r) \\approx {1 \over \sqrt{2\pi\sigma^2}e^{-r / (2\sigma^2)}}

  Args:
    displacement_or_metric:
      A function that computes the displacement or distance between two points.
    radii: An array of radii at which we would like to compute :math:`g(r)`.
    sigima: A float specifying the width of the approximating Gaussian.
    species: An optional array specifying the species of each particle. If
      species is None then we compute a single :math:`g(r)` for all particles,
      otherwise we compute one :math:`g(r)` for each species.
    eps: A small additive constant used to ensure stability if the radius is
      zero.

  Returns:
    A function `g_fn` that computes the pair correlation function for a
    collection of particles.

  :math:`g(r)` is calculated separately for each particle. For species=None, the
  output of `g_fn` is an array of shape (N, nr), where N is the number of 
  particles passed to `g_fn` and nr is the size of radii (the number of points 
  at which we calculate :math:`g(r)`. When species is specified, the output is a
  list of nspecies arrays, each of shape (N, nr), where nspecies is the number
  of unique species. If `gofr` is the output of `g_fn`, then gofr[si][i] gives
  the :math:`g(r)` for particle i considering only pair particles of species si.

  Note: when species is specified, the returned list is in the order of the 
  sorted unique species indices, not the order in which they appear. 
  
  """
  d = space.canonicalize_displacement_or_metric(displacement_or_metric)
  d = space.map_product(d)

  inv_rad = 1 / (radii + eps)

  def pairwise(dr, dim):
    return jnp.exp(-f32(0.5) * (dr - radii)**2 / sigma**2) * inv_rad**(dim - 1)
  pairwise = vmap(vmap(pairwise, (0, None)), (0, None))

  if species is None:
    def g_fn(R):
      dim = R.shape[-1]
      mask = 1 - jnp.eye(R.shape[0], dtype=R.dtype)
      g_R = jnp.sum(mask[:, :, jnp.newaxis] *
                    pairwise(d(R, R), dim), axis=(1,))
      if compute_average:
        g_R = average_pair_correlation_results(g_R, species)
      return g_R
  else:
    if not (isinstance(species, jnp.ndarray) and is_integer(species)):
      raise TypeError('Malformed species; expecting array of integers.')
    species_types = jnp.unique(species)
    def g_fn(R):
      dim = R.shape[-1]
      g_R = []
      mask = 1 - jnp.eye(R.shape[0], dtype=R.dtype)
      for s in species_types:
        Rs = R[species == s]
        mask_s = mask[:, species == s, jnp.newaxis]
        g_R += [jnp.sum(mask_s * pairwise(d(Rs, R), dim), axis=(1,))]
      if compute_average:
        g_R = average_pair_correlation_results(g_R, species)
      return g_R
  return g_fn


def pair_correlation_neighbor_list(
    displacement_or_metric: Union[DisplacementFn, MetricFn],
    box_size: Box,
    radii: Array,
    sigma: float,
    species: Array = None,
    dr_threshold: float = 0.5,
    eps: float = 1e-7,
    fractional_coordinates: bool=False,
    format: partition.NeighborListFormat=partition.Dense,
    compute_average: bool = False):
  """Computes the pair correlation function at a mesh of distances.

  The pair correlation function measures the number of particles at a given
  distance from a central particle. The pair correlation function is defined by 
  
  .. math::
    g(r) = <\sum_{i \\neq j}\delta(r - |r_i - r_j|)>
  
  We make the approximation,

  .. math::
    \delta(r) \\approx {1 \over \sqrt{2\pi\sigma^2}e^{-r / (2\sigma^2)}}

  This function uses neighbor lists to speed up the calculation.

  Args:
    displacement_or_metric: A function that computes the displacement or
      distance between two points.
    box_size: The size of the box containing the particles.
    radii: An array of radii at which we would like to compute :math:`g(r)`.
    sigima: A float specifying the width of the approximating Gaussian.
    species: An optional array specifying the species of each particle. If
      species is None then we compute a single :math:`g(r)` for all particles,
      otherwise we compute one :math:`g(r)` for each species.
    dr_threshold: A float specifying the halo size of the neighbor list.
    eps: A small additive constant used to ensure stability if the radius is
      zero.
    fractional_coordinates: Bool determining whether positions are stored in
      the unit cube or not.
    format: The format of the neighbor lists. Must be `Dense` or `Sparse`.

  Returns:
    A pair of functions: `neighbor_fn` that constructs a neighbor list (see
    `neighbor_list` in `partition.py` for details). `g_fn` that computes the
    pair correlation function for a collection of particles given their
    position and a neighbor list.

  :math:`g(r)` is calculated separately for each particle. For species=None, the
  output of `g_fn` is an array of shape (N, nr), where N is the number of 
  particles passed to `g_fn` and nr is the size of radii (the number of points 
  at which we calculate :math:`g(r)`. When species is specified, the output is a
  list of nspecies arrays, each of shape (N, nr), where nspecies is the number
  of unique species. If `gofr` is the output of `g_fn`, then gofr[si][i] gives
  the :math:`g(r)` for particle i considering only pair particles of species si.

  Note: when species is specified, the returned list is in the order of the
  sorted unique species indices, not the order in which they appear. 
  """
  metric = space.canonicalize_displacement_or_metric(displacement_or_metric)
  inv_rad = 1 / (radii + eps)
  def pairwise(dr, dim):
    return jnp.exp(-f32(0.5) * (dr - radii)**2 / sigma**2) * inv_rad**(dim - 1)

  neighbor_fn = partition.neighbor_list(displacement_or_metric,
                                        box_size,
                                        jnp.max(radii) + sigma,
                                        dr_threshold,
                                        format=format)

  if species is None:
    def g_fn(R, neighbor):
      N, dim = R.shape
      mask = partition.neighbor_list_mask(neighbor)
      if neighbor.format is partition.Dense:
        R_neigh = R[neighbor.idx]
        d = space.map_neighbor(metric)
        _pairwise = vmap(vmap(pairwise, (0, None)), (0, None))
        g_R = jnp.sum(mask[:, :, None] *
                      _pairwise(d(R, R_neigh), dim), axis=(1,))
        if compute_average:
          g_R = average_pair_correlation_results(g_R, species)
        return g_R
      elif neighbor.format is partition.Sparse:
        dr = space.map_bond(metric)(R[neighbor.idx[0]], R[neighbor.idx[1]])
        _pairwise = vmap(pairwise, (0, None))
        g_R = ops.segment_sum(mask[:, None] * _pairwise(dr, dim),
                              neighbor.idx[0],
                              N)
        if compute_average:
          g_R = average_pair_correlation_results(g_R, species)
        return g_R
      else:
        raise NotImplementedError('Pair correlation function does not support '
                                  'OrderedSparse neighbor lists.')

  else:
    if not (isinstance(species, jnp.ndarray) and is_integer(species)):
      raise TypeError('Malformed species; expecting array of integers.')
    species_types = jnp.unique(species)
    def g_fn(R, neighbor):
      N, dim = R.shape
      g_R = []
      mask = partition.neighbor_list_mask(neighbor)
      if neighbor.format is partition.Dense:
        neighbor_species = species[neighbor.idx]
        R_neigh = R[neighbor.idx]
        d = space.map_neighbor(metric)
        _pairwise = vmap(vmap(pairwise, (0, None)), (0, None))
        for s in species_types:
          mask_s = mask * (neighbor_species == s)
          g_R += [jnp.sum(mask_s[:, :, jnp.newaxis] *
                          _pairwise(d(R, R_neigh), dim), axis=(1,))]
      elif neighbor.format is partition.Sparse:
        neighbor_species = species[neighbor.idx[1]]
        dr = space.map_bond(metric)(R[neighbor.idx[0]], R[neighbor.idx[1]])
        _pairwise = vmap(pairwise, (0, None))
        for s in species_types:
          mask_s = mask * (neighbor_species == s)
          g_R += [ops.segment_sum(mask_s[:, None] *
                                  _pairwise(dr, dim), neighbor.idx[0], N)]
      else:
        raise NotImplementedError('Pair correlation function does not support '
                                  'OrderedSparse neighbor lists.')

      if compute_average:
        g_R = average_pair_correlation_results(g_R, species)
      return g_R
  return neighbor_fn, g_fn

def nball_unit_volume(spatial_dimension: int) -> float:
  """ Return the volume of a unit sphere in arbitrary dimensions
  """
  return jnp.power(jnp.pi, spatial_dimension / 2) / \
    jnp.exp( jsp.special.gammaln(spatial_dimension / 2 + 1))

def particle_volume(radii: Array, 
                    spatial_dimension: int, 
                    particle_count: Array = 1, 
                    species: Array = None) -> float:
  """ Calculate the volume of a collection of particles

  Args:
    radii: array of shape (n,) giving particle radii, where n can be 1, the
      number of species, or the number of particles depending on the values of
      particle_count and species. 
    spatial_dimension: int giving the spatial dimension
    particle_count: number of particles with each radii. Broadcastable to radii.
    species: list of particle species. If provided, this overrides 
      particle_count and radii is expected to give per-species radii
  
  Returns: the sum of the volume of all the particles
  """
  V_unit = nball_unit_volume(spatial_dimension)
  V_particle = V_unit * radii**spatial_dimension

  if species is not None:
    particle_count = jnp.bincount(species)

  return jnp.sum(particle_count * V_particle)

def volume_fraction(box: Box, 
                    radii: Array, 
                    spatial_dimension: int, 
                    particle_count: Array = 1, 
                    species: Array = None) -> float:
  """ Calculate the volume fraction

  See documentation for particle_volume for explanation of parameters
  """
  Vparticle = particle_volume(radii, spatial_dimension, particle_count, species)
  return Vparticle / quantity.volume(spatial_dimension, box)

def box_size_at_volume_fraction(volume_fraction: float,
                                radii: Array,
                                spatial_dimension: int,
                                particle_count: Array = 1,
                                species: Array = None) -> float:
  """ Calculate box_size to obtain a desired volume fraction

  See documentation for particle_volume for explanation of parameters
  """
  Vparticle = particle_volume(radii, spatial_dimension, particle_count, species)
  return jnp.power( Vparticle / volume_fraction, 1 / spatial_dimension)

def box_size_at_number_density(particle_count: int,
                               number_density: float,
                               spatial_dimension: int) -> float:
  return jnp.power(particle_count / number_density, 1 / spatial_dimension)


def box_from_parameters(a: float, b: float, c: float,
                        alpha: float, beta: float, gamma: float) -> Box:
  alpha = alpha * jnp.pi / 180
  beta = beta * jnp.pi / 180
  gamma = gamma * jnp.pi / 180
  yy = b * jnp.sin(gamma)
  xy = b * jnp.cos(gamma)
  xz = c * jnp.cos(beta)
  yz = (b * c * jnp.cos(alpha) - xy * xz) / yy
  zz = jnp.sqrt(c**2 - xz**2 - yz**2)
  return jnp.array([
      [a, xy, xz],
      [0, yy, yz],
      [0, 0,  zz]
  ])

def bulk_modulus(elastic_tensor: Array) -> float:
  return jnp.einsum('iijj->', elastic_tensor) / elastic_tensor.shape[0] ** 2


@dataclasses.dataclass
class PHopState:
    position_buffer: jnp.ndarray
    phop: jnp.ndarray

InitFn = Callable[[Array], PHopState]
ApplyFn = Callable[[PHopState, Array], PHopState]
PHopCalculator = Tuple[InitFn, ApplyFn]

def phop(displacement: DisplacementFn, window_size: int) -> PHopCalculator:
  """Computes the phop indicator of rearrangements.

  phop is an indicator function that is effective at detecting when particles
  in a quiescent system have experienced a rearrangement. Qualitatively, phop
  measures when the average position of a particle has changed significantly.

  Formally, given a window of size :math:`\Delta t` we two averages before and after
  the current time,

  .. math::
    E_A[f] = E_{t\in[t - \Delta t / 2, t]}[f(t)]
    E_B[f] = E_{t\in[t, t + \Delta t / 2]}[f(t)].

  In terms of these expectations, phop is given by,

  .. math::
    phop = \sqrt{E_A[(R_i(t) - E_B[R_i(t)])^2]E_B[(R_i(t) - E_A[R_i(t)])^2]}

  phop was first introduced in Candelier et al. [#candelier]_

  Args:
    displacement: A function that computes displacements between pairs of
      particles. See `spaces.py` for details.
    window_size: An integer specifying the number of positions that constitute
      the window.

  Returns:
    A pair of functions, `(init_fn, update_fn)` that initialize the state of a
    phop measurement and update the state of a phop measurement to include new
    positions.

  .. rubric:: References
  .. [#candelier] R. Candelier et al.
    "Spatiotemporal Hierarchy of Relaxation Events, Dynamical Heterogeneities,
    and Structural Reorganization in a Supercooled Liquid"
    Physical Review Letters 105, 135702 (2010).
  """

  half_window_size = window_size // 2
  displacement = space.map_bond(displacement)

  def init_fn(position: Array) -> PHopState:
    position_buffer = jnp.tile(position, (window_size, 1, 1))
    assert position_buffer.shape == ((window_size,) + position.shape)
    return PHopState(position_buffer, jnp.zeros((position.shape[0],)))  # pytype: disable=wrong-arg-count

  def update_fn(state: PHopState, position: Array) -> PHopState:
    # Compute phop.
    a_pos = state.position_buffer[:half_window_size]
    a_mean = jnp.mean(a_pos, axis=0)
    b_pos = state.position_buffer[half_window_size:]
    b_mean = jnp.mean(b_pos, axis=0)

    phop = jnp.sqrt(jnp.mean((a_pos - b_mean) ** 2 * (b_pos - a_mean) ** 2,
                             axis=(0, 2)))

    # Unwrap position.
    buff = state.position_buffer
    position = displacement(position, buff[-1]) + buff[-1]

    # Add position to the list.
    buff = jnp.concatenate((buff, position[jnp.newaxis, :, :]))[1:]

    return PHopState(buff, phop)  # pytype: disable=wrong-arg-count

  return init_fn, update_fn
